package alterrs.jbct;

import static org.objectweb.asm.Opcodes.ACC_ABSTRACT;
import static org.objectweb.asm.Opcodes.ACC_NATIVE;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.objectweb.asm.Opcodes;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.FieldNode;
import org.objectweb.asm.tree.IincInsnNode;
import org.objectweb.asm.tree.IntInsnNode;
import org.objectweb.asm.tree.JumpInsnNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.LookupSwitchInsnNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.MultiANewArrayInsnNode;
import org.objectweb.asm.tree.TableSwitchInsnNode;
import org.objectweb.asm.tree.TryCatchBlockNode;
import org.objectweb.asm.tree.TypeInsnNode;
import org.objectweb.asm.tree.VarInsnNode;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonObject;

import alterrs.commons.Attributes;
import alterrs.commons.Constants;
import alterrs.jbct.utility.ClassCollection;
import alterrs.jbct.utility.FieldID;
import alterrs.jbct.utility.ID;
import alterrs.jbct.utility.Transformer;
import rs.emulate.lynx.Lynx;

/**
 * @author Lazaro
 */
public class JBCT {

	public static JBCT ctx = null;

	public static boolean keepNames = false;
	
	public static void main(String[] args) {
		try {
			System.out.println("//***************************************************************************\\\\");
			System.out.println("   AlterRS [Java Bytecode Transformer]");
			System.out.println("   Copyright (C) AlterRS Development Team 2011");
			System.out.println("   Running " + System.getProperty("os.name") + " on a(n) " + System.getProperty("os.arch") + " architecture");
			System.out.println("   Java version is " + System.getProperty("java.version"));
			System.out.println("\\\\***************************************************************************//");
			System.out.println();

			ctx = new JBCT();
			ctx.run();
		} catch (Throwable e) {
			System.err.println("Failed to run JBCT!");
			e.printStackTrace();
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e2) {
			}

			System.exit(1);
		}
	}

	public Attributes atr = new Attributes();

	public ClassCollection module;

	private void run() throws Throwable {

		System.out.println("Loading properties...");
		List<String> transformerNames = new ArrayList<>();
		Gson gson = new GsonBuilder().setLenient().create();
		try (BufferedReader reader = new BufferedReader(new InputStreamReader(
				new FileInputStream(Constants.PROPS)))) {

			JsonObject obj = gson.fromJson(reader, JsonObject.class);
			
			Constants.tmode = TMode.valueOf(obj.get("tmode").getAsInt());
			Constants.dmode = DMode.valueOf(obj.get("dmode").getAsInt());
			Constants.secret = obj.get("secret").getAsString();
			Constants.vector = obj.get("vector").getAsString();
			Constants.packages = obj.get("packages").getAsBoolean();
			Constants.defPackage = obj.get("package").getAsString();
			Constants.mapping = obj.get("mapping").getAsBoolean();
			
			JsonArray arr = obj.get("transformers").getAsJsonArray();
			for (int i = 0; i < arr.size(); i++)
				transformerNames.add(arr.get(i).getAsString());
			
			reader.close();
		} catch (Exception e) {
			System.err.println("Error loading properties!");
			e.printStackTrace();
		}
		System.out.println("Loaded properties!");
		System.out.println();
		
		if (Constants.tmode.equals(TMode.FETCH)) {
			System.out.println("Fetching lastest gamepack...");
			Lynx lynx = new Lynx(true);
			lynx.init();
			String ver = lynx.run();
			Constants.dir = Constants.BASE + ver;
			System.out.println("Fetched latest gamepack! (" + ver + ")");
		} else if (Constants.tmode.equals(TMode.DECRYPT)) {
			System.out.println("Decrypting gamepack...");
			Constants.dir = Constants.BASE + "local";
			Lynx lynx = new Lynx(false);
			lynx.localDecrypt(Constants.secret, Constants.vector);
			System.out.println("Decrypted gamepack!");
		} else if (Constants.tmode.equals(TMode.DEOB)) {
			Constants.dir = Constants.BASE + "local";
		}

		
		System.out.println();
		System.out.println("Loadeding transformers...");
		List<Transformer> transformers = new LinkedList<Transformer>();
		for (String tName : transformerNames) {
			Class<?> tClass = Class.forName(tName);

			Transformer t = (Transformer) tClass.newInstance();

			transformers.add(t);
		}
		System.out.println("Loaded " + transformers.size() + " transformers!");
		System.out.println();
		
		System.out.println("Loading classes...");
		module = new ClassCollection(new URL("jar:file:" + Constants.dir + "/client.jar!/"));
		module.load();
		System.out.println("Loaded " + module.getClasses().size() + " client classes!");
		System.out.println();
		
		if (Constants.mapping) {
			File file = new File(Constants.dir + "/mapping.txt");
			if (file.exists()) {
				System.out.println("Remmaping from file...");
				try (BufferedReader reader = new BufferedReader(new FileReader(file))) {
					reader.lines().forEach((String s) -> {
						if (s.charAt(0) == 'c') {
							String[] ss = s.substring(2).split("!");
							module.getRefactorer().refactorClass(new ID(ss[0]), ss[1]);
						} else if (s.charAt(0) == 'f') {
							String[] ss = s.substring(2).split("!");
							String[] sss = ss[0].split("#");
							module.getRefactorer().refactorField(new FieldID(sss[0], sss[1], sss[2]), ss[1]);
						}
					});
				}
				module.transform();
				
				System.out.println("Remapped from file!");
				System.out.println();
			}
		}
		
		
		transform(module, transformers);

		module.transform();
		
		System.out.println("Sorting Variables...");
		module.sortVars();
		System.out.println("Sorted Variables!");
		System.out.println();
		
		boolean jsrs = false;
		
		classloop: for (ClassNode n : module.getClasses().values()) {
			for (Object mo : n.methods) {
				MethodNode m = (MethodNode) mo;
				
				for (AbstractInsnNode ain : m.instructions.toArray()) {
					if (ain.getOpcode() == Opcodes.JSR) {
						jsrs = true;
						break classloop;
					}
				}
			}
		}
		
		if (jsrs) {
			System.out.println("Inlining JSRs...");
			module.inlineJSR();
			System.out.println("Inlined JSRs!");
			System.out.println();
		}

		System.out.println("Saving client...");
		module.save(new File(Constants.dir + "/clientDeob.jar"));
		System.out.println("Saved client!");
		System.out.println();
		
		if (Constants.dmode.equals(DMode.JODE)) {
			System.out.println("Decompiling...");
			
			jode.decompiler.Main.main(new String[] { "--dest", Constants.dir + "/source", Constants.dir + "/clientDeob.jar" });
			
			if (!Files.exists(Paths.get(Constants.dir + "/source/com/")))
				Files.createDirectories(Paths.get(Constants.dir + "/source/com/"));
			
			Files.copy(Paths.get(Constants.LOADER), Paths.get(Constants.dir + "/source/com/RS3Applet.java"));
		} else if (Constants.dmode.equals(DMode.FERNFLOWER)) {
			System.out.println("Decompiling...");
			
			org.jetbrains.java.decompiler.main.decompiler.ConsoleDecompiler.main(new String[] { "-mpm=1000", Constants.dir + "/clientDeob.jar", Constants.dir + "/source" });
			
			if (!Files.exists(Paths.get(Constants.dir + "/source/com/")))
				Files.createDirectories(Paths.get(Constants.dir + "/source/com/"));
			
			Files.copy(Paths.get(Constants.LOADER), Paths.get(Constants.dir + "/source/com/RS3Applet.java"));
			
			System.out.println("Decompiled!");
		}
	}

	private void transform(ClassCollection module, List<Transformer> transformers) {
		trans: for (Transformer t : transformers) {
			try {
				System.out.println("Applying " + t.getClass().getName() + "...");

				t.transform(module);
				for (ClassNode c : module.getClasses().values()) {
					if (t.finished()) {
						t.onFinish();
						continue trans;
					}

					t.visitClass(c);

					for (Object fo : c.fields) {
						if (t.finished()) {
							break;
						}

						t.visitField(c, (FieldNode) fo);
					}

					for (Object mo : c.methods) {
						if (t.finished()) {
							break;
						}

						MethodNode m = (MethodNode) mo;
						
						if ((m.access & ACC_ABSTRACT) != 0
								|| (m.access & ACC_NATIVE) != 0) {
							t.visitNativeMethod(c, m);
							continue;
						}

						t.visitMethod(c, m);

						AbstractInsnNode[] insns = m.instructions.toArray();
						for (AbstractInsnNode insn : insns) {
							if (insn == null || !m.instructions.contains(insn)) {
								continue;
							}
							if (t.finished()) {
								break;
							}

							t.visitInsn(c, m, insn);

							if (insn instanceof IntInsnNode) {
								t.visitIntInsn(c, m, (IntInsnNode) insn);
							}
							if (insn instanceof VarInsnNode) {
								t.visitVarInsn(c, m, (VarInsnNode) insn);
							}
							if (insn instanceof TypeInsnNode) {
								t.visitTypeInsn(c, m, (TypeInsnNode) insn);
							}
							if (insn instanceof FieldInsnNode) {
								t.visitFieldInsn(c, m, (FieldInsnNode) insn);
							}
							if (insn instanceof MethodInsnNode) {
								t.visitMethodInsn(c, m, (MethodInsnNode) insn);
							}
							if (insn instanceof JumpInsnNode) {
								t.visitJumpInsn(c, m, (JumpInsnNode) insn);
							}
							if (insn instanceof LdcInsnNode) {
								t.visitLdcInsn(c, m, (LdcInsnNode) insn);
							}
							if (insn instanceof IincInsnNode) {
								t.visitIincInsn(c, m, (IincInsnNode) insn);
							}
							if (insn instanceof TableSwitchInsnNode) {
								t.visitTableSwitchInsn(c, m,
										(TableSwitchInsnNode) insn);
							}
							if (insn instanceof LookupSwitchInsnNode) {
								t.visitLookupSwitchInsn(c, m,
										(LookupSwitchInsnNode) insn);
							}
							if (insn instanceof MultiANewArrayInsnNode) {
								t.visitMultiANewArrayInsn(c, m,
										(MultiANewArrayInsnNode) insn);
							}
						}

						for (Object tcbo : m.tryCatchBlocks) {
							if (t.finished()) {
								break;
							}
							TryCatchBlockNode tcb = (TryCatchBlockNode) tcbo;

							t.visitTryCatchBlock(c, m, tcb);
						}
					}
				}
				if (t.finished() || t.forceFinish()) {
					t.onFinish();
					System.out.println();
				}
			} catch (Exception e) {
				System.err.println("Exception on transformer : " + t);
				e.printStackTrace();
			}
		}
	}
}
