package alterrs.jbct.deob;

import static org.objectweb.asm.Opcodes.ATHROW;
import static org.objectweb.asm.Opcodes.ICONST_1;
import static org.objectweb.asm.Opcodes.IFLE;
import static org.objectweb.asm.Opcodes.IF_ICMPLE;
import static org.objectweb.asm.Opcodes.IF_ICMPNE;
import static org.objectweb.asm.Opcodes.ILOAD;
import static org.objectweb.asm.Opcodes.NEW;
import static org.objectweb.asm.Opcodes.RETURN;

import java.util.HashSet;
import java.util.Set;

import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.JumpInsnNode;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.LineNumberNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.TypeInsnNode;
import org.objectweb.asm.tree.VarInsnNode;

import alterrs.jbct.utility.ASMUtil;
import alterrs.jbct.utility.AbstractTransformer;

/**
 * @author Lazaro Brito
 * @authro Kyle Friz - Fixed the logic
 */
public class ConditionalDeobfuscator extends AbstractTransformer {
	private int count = 0;

	public ConditionalDeobfuscator() {
		super(true);
	}

	@Override
	public void visitVarInsn(ClassNode c, MethodNode m, VarInsnNode n) {
		if (n.getOpcode() == ILOAD) {
			AbstractInsnNode n1 = ASMUtil.previousRealInsn(n.getPrevious());// comparator
			if (n1 == null)
				return;
			
			AbstractInsnNode n0 = ASMUtil.previousRealInsn(n1.getPrevious());// var
			if (n0 == null)
				return;
			
			AbstractInsnNode n2 = ASMUtil.nextRealInsn(n.getNext());// constant
			if (n2 == null)
				return;

			AbstractInsnNode n3 = ASMUtil.nextRealInsn(n2.getNext());// comparator
			if (n3 == null)
				return;

			AbstractInsnNode n4 = ASMUtil.nextRealInsn(n3.getNext());// fail
			if (n4 == null)
				return;

			if (!isLE(n1) || !isConst(n2) || !isCMPLE(n3) || !isNew(n4))
				return;
			
			Set<AbstractInsnNode> del = new HashSet<AbstractInsnNode>();
			AbstractInsnNode ain = n3;
			while (true) {
				ain = ain.getNext();
				del.add(ain);
				
				if (ain.getOpcode() == ATHROW)
					break;
			}
			
			for (AbstractInsnNode i : del) {
				m.instructions.remove(i);
			}
			
			for (AbstractInsnNode a : m.instructions.toArray()) {
				if (a.getOpcode() == RETURN) {
					if (a.getPrevious() instanceof LineNumberNode) {
						if (a.getPrevious().getPrevious() instanceof LabelNode) {
							m.instructions.set(n3, new JumpInsnNode(IF_ICMPNE, (LabelNode) a.getPrevious().getPrevious()));
						}
					} else if (a.getPrevious() instanceof LabelNode) {
						m.instructions.set(n3, new JumpInsnNode(IF_ICMPNE, (LabelNode) a.getPrevious()));
					}
				}
			}

			count++;
			
			m.instructions.remove(n0);
			m.instructions.remove(n1);
		}
	}

	private boolean isConst(AbstractInsnNode n) {
		switch (n.getOpcode()) {
		case ICONST_1:
			return true;
		}
		return false;
	}

	private boolean isLE(AbstractInsnNode n) {
		switch (n.getOpcode()) {
		case IFLE:
			return true;
		}
		return false;
	}
	
	private boolean isCMPLE(AbstractInsnNode n) {
		switch (n.getOpcode()) {
		case IF_ICMPLE:
			return true;
		}
		return false;
	}
	
	private boolean isNew(AbstractInsnNode n) {
		if (n.getOpcode() == NEW)
			return ((TypeInsnNode) n).desc.equals("java/lang/IllegalStateException");
			
		return false;
	}

	@Override
	public void onFinish() {
		System.out.println(" ^	Repaired " + count + " redundant conditionals!");
	}
}
