package alterrs.jbct.deob;

import static org.objectweb.asm.Opcodes.ALOAD;
import static org.objectweb.asm.Opcodes.ASTORE;
import static org.objectweb.asm.Opcodes.DUP;
import static org.objectweb.asm.Opcodes.MONITORENTER;

import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.VarInsnNode;

import alterrs.jbct.utility.ASMUtil;
import alterrs.jbct.utility.AbstractTransformer;

/**
 * @author Lazaro Brito
 */
public class MonitorDeobfuscator extends AbstractTransformer {
	private int count = 0;

	public MonitorDeobfuscator() {
		super(true);
	}

	@Override
	public void visitInsn(ClassNode c, MethodNode m, AbstractInsnNode n) {
		if (n.getOpcode() == MONITORENTER) {
			AbstractInsnNode astore = ASMUtil.previousRealInsn(n.getPrevious());
			if (astore == null) {
				return;
			}

			AbstractInsnNode dup = ASMUtil.previousRealInsn(astore
					.getPrevious());
			if (dup == null) {
				return;
			}

			if (astore.getOpcode() == ASTORE && dup.getOpcode() == DUP) {
				m.instructions.remove(astore);
				m.instructions.set(dup, astore);

				m.instructions.insert(astore, new VarInsnNode(ALOAD,
						((VarInsnNode) astore).var));

				count++;
			}
		}
	}

	@Override
	public void onFinish() {
		System.out.println(" ^	Repaired " + count + " monitor blocks!");
	}
}
