package alterrs.jbct.deob;

import static org.objectweb.asm.Opcodes.ACONST_NULL;
import static org.objectweb.asm.Opcodes.INVOKESPECIAL;

import org.objectweb.asm.Type;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;

import alterrs.jbct.utility.ASMUtil;
import alterrs.jbct.utility.AbstractTransformer;

/**
 * 
 * @author Kyle Friz
 * @since  Apr 24, 2016
 */
public class SuperInvokeDeobfuscator extends AbstractTransformer {
	private int count = 0;

	public SuperInvokeDeobfuscator() {
		super(true);
	}

	@Override
	public void visitMethodInsn(ClassNode c, MethodNode m, MethodInsnNode n) {
		if (m.name.equals("<init>") && n.name.equals("<init>") &&  n.getOpcode() == INVOKESPECIAL) {
			Type[] types = Type.getArgumentTypes(n.desc);
			
			if (types.length == 5) {
				if (types[0].getSort() == Type.OBJECT
						&& types[1].getSort() == Type.INT
						&& types[2].getSort() == Type.BOOLEAN
						&& types[3].getSort() == Type.BOOLEAN
						&& types[4].getSort() == Type.OBJECT) {
					
					AbstractInsnNode aconst_null = ASMUtil.previousRealInsn(n.getPrevious());
					
					if (aconst_null.getOpcode() == ACONST_NULL) {
						m.instructions.remove(aconst_null);
						
						Type[] newTypes = new Type[4];
						System.arraycopy(types, 0, newTypes, 0, newTypes.length);

						n.desc = Type.getMethodDescriptor(Type.getReturnType(n.desc), newTypes);

						count++;
					}
				}
			}
		}
	}

	@Override
	public void onFinish() {
		System.out.println(" ^	Repaired " + count + " super invokes!");
	}
}
