package alterrs.jbct.ref;

import static org.objectweb.asm.Opcodes.ACC_INTERFACE;
import static org.objectweb.asm.Opcodes.ACC_NATIVE;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.MethodNode;

import alterrs.commons.Constants;
import alterrs.jbct.JBCT;
import alterrs.jbct.utility.AbstractTransformer;
import alterrs.jbct.utility.ClassCollection;
import alterrs.jbct.utility.ID;

/**
 * @author Lazaro Brito
 */
public class ClassNameDeobfuscator extends AbstractTransformer {
	private int interfaceId = 1;
	private int nCount = 0;
	private Map<ID, Integer> classIds = new HashMap<ID, Integer>();
	private int cCount = 0;

	public ClassNameDeobfuscator() {
		super(true);
		JBCT.ctx.atr.set("ignoredClasses", new HashSet<ID>());
	}

	@Override
	public void transform(ClassCollection cc) {
		for (int level = 0; level < 32; level++) {
			classLoop: for (ClassNode c : cc.getClasses().values()) {
				if (cc.getLevel(c) == level) {
					String cName = c.name.substring(
							c.name.lastIndexOf("/") + 1, c.name.length());
					boolean isDefaultPackage = (c.name.indexOf("/") == -1);
					if (cName.length() <= 3 && !cName.equals("Npc") && !cName.equals("GPI")
							&& !containsNativeMethods(c)
							&& JBCT.ctx.module.getRefactorer().getClass(
									new ID(c)) == null) {
						
						//System.out.println(c.name);
						String name;
						if ((c.access & ACC_INTERFACE) != 0) {
							interfaceId++;
							
							if (JBCT.keepNames)
								name = "Interface_" + c.name;
							else
								name = "Interface" + interfaceId;
						} else {
							name = name(c);
							cCount++;
						}
						//name = c.name;
						if (isDefaultPackage && Constants.packages) {
							JBCT.ctx.module.getRefactorer().refactorClass(
								new ID(c), Constants.defPackage + name);
						} else {
							JBCT.ctx.module.getRefactorer().refactorClass(
									new ID(c), c.name.substring(0, c.name.lastIndexOf("/") + 1)+name);
						}
					} else {
						if (!containsNativeMethods(c)) {
							if (isDefaultPackage && Constants.packages)
								JBCT.ctx.module.getRefactorer().refactorClass(new ID(c), Constants.defPackage + cName);
						} else {
							nCount++;
						}
						
						for (Object mo : c.methods) {
							MethodNode m = (MethodNode) mo;
							if (m.name.equals("<init>")) {
								for (AbstractInsnNode n : m.instructions.toArray()) {
									if (n instanceof LdcInsnNode) {
										if (((LdcInsnNode) n).cst
												.equals("sw3d")) {
											JBCT.ctx.atr.<Set<ID>> get(
													"ignoredClasses").add(
													new ID(c));
											continue classLoop;
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}

	@Override
	public void onFinish() {
		System.out.println(" ^	Refactored " + cCount + " classes and " + (interfaceId - 1) + " interfaces!");
		System.out.println(" ^	Skipped " + nCount + " native classes!");
	}

	private int id(ClassNode c) {
		if (c.superName == null) {
			return 1;
		} else {
			Integer i = classIds.get(new ID(c.superName));
			if (i == null) {
				i = 0;
			}
			i++;
			classIds.put(new ID(c.superName), i);
			return i;
		}
	}

	private String name(ClassNode c) {
		int id = id(c);
		int level = JBCT.ctx.module.getLevel(c);

		if (level == 1) {
			if (JBCT.keepNames)
				return "Class_" + c.name;
			
			return "Class" + id;
		} else {
			StringBuilder name = new StringBuilder();
			String superName = JBCT.ctx.module.getRefactorer().getClass(
					new ID(c.superName));
			if (superName != null) {
				name.append(superName.substring(superName.lastIndexOf("/") + 1, superName.length()));
			} else {
				name.append(c.superName.substring(
						c.superName.lastIndexOf("/") + 1, c.superName.length()));
			}
			
			if (JBCT.keepNames)
				name.append("_").append(c.name);
			
			name.append("_Sub").append(id);
			return name.toString();
		}
	}

	private boolean containsNativeMethods(ClassNode c) {
		for (Object mo : c.methods) {
			MethodNode m = (MethodNode) mo;
			if ((m.access & ACC_NATIVE) != 0) {
				return true;
			}
		}
		return false;
	}
}